!     ------------------------------------------------------------------
!
!     MOSFET.f95
!
!
!     Created on Mon Apr 03 50:09:51 2007
!     Copyright (c) 2007 James K Beard. All rights reserved.
!
!
!     ------------------------------------------------------------------

module MOSFETs	!Support model and v-i curve family for IGFET
implicit none
!Elementary model
!Triode region model is ID = K*(2*(VG-VTransistion)*VD - VD^2)
!Constant current model is ID = K*(VG-VTransistion)^2+VD/RDrain
!The transition between these is when VD = VG - VTransistion, or ID = K*VD^2
!The intersection of these lines is always included in the output table

!Public transistor parameters, with defaults
real(kind(0.d0))::VTransition=2d0,K_conductance_parameter=0.8d-3,V_DS_Early=100d0

!User entry points
!## real(kind(0.d0)) function ID_vs_VDS(VDr,VGt)	!Drain current vs. Drain voltage, Gate voltage
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion			Transistion gate-source voltage, volts, default 2.0 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 8000 uMhos/Ohm
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
!Arguments
! VDr			Drain to source voltage, Volts
! VGt			Gate to source voltage, Volts
!Returned value:  Drain current, in amperes

!## real(kind(0.d0)) function MOSFET_Transconductance(VDr,VGt)	!Drain current vs. Drain voltage, Gate voltage
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion			Transistion gate-source voltage, volts, default 2.0 Volts
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 8000 uMhos/Ohm
!Arguments
! VDr			Drain to source voltage, Volts
! VGt			Gate to source voltage, Volts
!Returned value:  Transconductance, in mhos

!## real(kind(0.d0)) function MOSFET_GConductance(VDr,VGt)	!Gate conductance vs. Drain voltage, Gate voltage
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion			Transistion gate-source voltage, volts, default 2.0 Volts
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 8000 uMhos/Ohm
!Arguments
! VDr			Drain to source voltage, Volts
! VGt			Gate to source voltage, Volts
!Returned value:  Drain conductance in mhos

!## subroutine generate_vi_csv(lunout,fname,VGDelta,NVGate,VDDelta,VDMax)
!Inputs and outputs follow subroutine definition, immediately below
!------------------------------------------------
contains
!--------------------------------------------------
subroutine generate_vi_csv(lunout,fname,VGDelta,NVGate,VDDelta,VDMax)
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion		Transistion gate-source voltage, volts, default 2.0 Volts
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 800 uMhos/Ohm
!Arguments
! lunout	Logical unit for output
! fname		file name, without extent
! VGDelta	Granularity of gate voltage for output files
! NVGate	Number of gate voltages to plot (determines number of v-i curves)
! VDelta	Granularity of drain voltages (determines number of rows of output)
! VDMax		Maximum drain voltage to plot
integer,intent(IN)::lunout,NVGate
character(len=*),intent(IN)::fname

real(kind(0d0)),intent(IN)::VGDelta,VDDelta,VDMax
real(kind(0.d0))::VDrain,VDIntercept1,VDInterceptN,VDDelta_start

integer::i	!For array generators
real(kind(0.d0)),dimension(NVGate+1)::VGTitle		!For titleline
real(kind(0.d0)),dimension(NVGate+1)::VGArray,IDArray	!For model_line(VD)

  VGTitle=VTransition+(/(i,i=0,NVGate)/)*VGDelta	!For title line
  VGArray=VTransition+(/(i,i=0,NVGate)/)*VGDelta	!For model_line

  open(lunout,file=trim(adjustl(fname))//".csv")

!Write column headings
!Column headings are:
!VD V    Bound    1.0 V    2.0 V     3.0 V ...
  write(lunout,1000) VGTitle	!Output in V

  VDIntercept1=VGDelta		!Find first transition, for first gate voltage over cutoff

  VDInterceptN=NVGate*VGDelta	!Find last transition, for highest gate voltage

  VDDelta_start=VDIntercept1/10d0	!Plot granularity is 10% of the first one

!Do a fine-grained plot through the triode region
  VDrain=0.d0
  do
    call model_line(VDrain)
    VDrain=VDrain+VDDelta_start		!For each Gate current
    if(VDrain >= VDInterceptN) exit	!Go to coarser resolution for constant current region
  end do

  VDrain=0.d0
  do
    VDrain=VDrain+VDdelta
    if(VDrain < VDInterceptN) cycle	!Skip if not past the intersections
    if(VDrain > VDmax) exit		!Done when past the limit
    call model_line(VDrain)
  end do

  close(lunout)

1000 format("VD V,Bound"20(:",",f4.1," V"))

contains
!------------------------------------
subroutine model_line(VD)
!Rows are:
!<VD, Volts>     <Boundary of triode region>     <ID, mA for VDS=1V> ...
!Boundary value is discontinued after the parabola crosses the last v-i curve
real(kind(0.d0)),intent(IN)::VD
integer::i
character(len=1024)::output_line
real(kind(0.d0))::Channel_Length

  Channel_Length=1.d0+VD/V_DS_Early

  do i=0,NVGate
    IDArray(i+1)=ID_vs_VDS(VD,VGArray(i+1))
  end do

  if(VD <= VDInterceptN) then
!    write(lunout,1000) VD,K_conductance_parameter*VD**2*1d3,IDArray*1d3	!Output in mA
    write(output_line,1000) VD,K_conductance_parameter*VD**2*Channel_Length*1d3,	&
      IDArray*1d3	!Output in mA
    call write_output(output_line)
  else
!    write(lunout,1010) VD,IDArray*1d3
    write(output_line,1010) VD,IDArray*1d3
    call write_output(output_line)
  end if

1000 format(g15.6,20(:",",g15.6))
1010 format(g15.6,",",20(:",",g15.6,:))
end subroutine model_line
!----------------------------------------------
subroutine write_output(line_of_text)		!Strip out blanks and write to outlun
character(len=1024),intent(IN)::line_of_text
character(len=1024)::work
integer::position

  work=trim(adjustl(line_of_text))	!Start by eliminating leading blanks
  do					!Strip out blanks, one by one
    position=index(trim(work)," ")	!Find blank, if any
    if(position == 0) exit		!If none, we are done
    work=work(1:position-1)//work(position+1:len_trim(work))
  end do

  write(lunout,1000) trim(work)

1000 format(a)
end subroutine write_output
!----------------------------------------------
end subroutine generate_vi_csv
!----------------------------------------------
real(kind(0.d0)) function ID_vs_VDS(VDr,VGt)	!Drain current vs. Drain voltage, Gate voltage
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion			Transistion gate-source voltage, volts, default 2.0 Volts
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 8000 uMhos/Ohm
!Arguments
! VDr			Drain to source voltage, Volts
! VGt			Gate to source voltage, Volts
!Returned value:  Drain current, in amperes

real(kind(0.d0)),intent(IN)::VDr,VGt
real(kind(0.d0))::Channel_Length
!This is the main engine of the module
!  ITriode=K_conductance_parameter*(2.d0*(VGt-VTransistion)*VDr-VDr**2)*Channel_Length
!  ICC=K_conductance_paramter*(VGt-VTransistion)**2*Channel_Length

  Channel_Length=1.d0+VDr/V_DS_Early

  if(VDr <= 0.d0 .OR. VGt <= VTransition) then
    ID_vs_VDS=0.d0
  elseif(VDr <= VGt-VTransition) then
    ID_vs_VDS=K_conductance_parameter*(2.d0*(VGt-VTransition)*VDr-VDr**2)*Channel_Length
  else
    ID_vs_VDS=K_conductance_parameter*(VGt-VTransition)**2*Channel_Length
  end if

end function ID_vs_VDS
!----------------------------------------
real(kind(0.d0)) function MOSFET_Transconductance(VDr,VGt)	!Transconductance vs. Drain voltage, Gate voltage
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion			Transistion gate-source voltage, volts, default 2.0 Volts
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 8000 uMhos/Ohm
!Arguments
! VDr			Drain to source voltage, Volts
! VGt			Gate to source voltage, Volts
!Returned value:  Transconductance, in mhos
real(kind(0.d0)),intent(IN)::VDr,VGt
real(kind(0.d0))::Channel_Length

  Channel_Length=1.d0+VDr/V_DS_Early

  if(VDr <= 0.d0 .OR. VGt <= VTransition) then
    MOSFET_Transconductance=0.d0
  elseif(VDr <= VGt-VTransition) then
    MOSFET_Transconductance=2.d0*K_conductance_parameter*VDr*Channel_Length
  else
    MOSFET_Transconductance=2.d0*K_conductance_parameter*(VGt-VTransition)*Channel_Length
  end if

end function MOSFET_Transconductance
!----------------------------------------
real(kind(0.d0)) function MOSFET_GConductance(VDr,VGt)	!Gate conductance vs. Drain voltage, Gate voltage
!INPUTS
!Transistor parameters, as global variables in this module
! VTransistion			Transistion gate-source voltage, volts, default 2.0 Volts
! V_DS_Early			Intercept of i=v curves for constant curreng region;, default 100 Volts
! K_Conductance_Parameter	Amps per Ohm squared, default 8000 uMhos/Ohm
!Arguments
! VDr			Drain to source voltage, Volts
! VGt			Gate to source voltage, Volts
!Returned value:  Drain conductance in mhos

real(kind(0.d0)),intent(IN)::VDr,VGt
!This is the main engine of the module
!  ITriode=K_conductance_parameter*(2.d0*(VGt-VTransistion)*VDr-VDr**2)
!  ICC=K_conductance_paramter*(VGt-VTransistion)**2*Channel_Length
real(kind(0.d0))::Channel_Length

  Channel_Length=1.d0+VDr/V_DS_Early

  if(VDr <= 0.d0 .OR. VGt <= VTransition) then
    MOSFET_GConductance=0.d0
  elseif(VDr <= VGt-VTransition) then
    MOSFET_GConductance=K_conductance_parameter*(2.d0*(VGt-VTransition-VDr)*Channel_Length	&
      +(2.d0*(VGt-VTransition)*VDr-VDr**2)/V_DS_Early)
  else
    MOSFET_GConductance=K_conductance_parameter*(VGt-VTransition)**2/V_DS_Early
  end if

end function MOSFET_GConductance
!----------------------------------------
end module MOSFETs
